/*
 *  Copyright (c) by Jaroslav Kysela <perex@jcu.cz>
 */

#define ULTRA_MAIN_OBJECT_FILE
#include "driver.h"

#if 0
#define DEBUG_ACCESS( d, m ) printk( ">>" d "[%i]\n", m )
#else
#define DEBUG_ACCESS( d, m )
#endif

int ultra_major = ULTRACFG_MAJOR;
int ultra_config_dma_static = 0;
#ifdef MODULE_PARM              /* hey - we have new 2.1.18+ kernel... */
MODULE_AUTHOR( "Jaroslav Kysela <perex@jcu.cz>" );
MODULE_DESCRIPTION( "Ultra Sound Low-level driver for soundcards." );
MODULE_SUPPORTED_DEVICE( "sound" );
MODULE_PARM( ultra_major, "i" );
MODULE_PARM_DESC( ultra_major, "Major # for Ultra driver. (14 by default)" );
MODULE_PARM( ultra_config_dma_static, "i" );
MODULE_PARM_DESC( ultra_config_dma_static, "Enable static DMA memory allocation. (off by default)" );
#endif

#define VERIFY_CARD( idx, minor ) \
  ( ultra_verify_map[ idx ] && !(ultra_cards_bitmap & (1 << ((minor >> 4) & 0x07))) )

#define ULTRA_MINOR_OSS		0x10

static char *ultra_comment_map[ 32 ];
static int ultra_verify_map[ 32 ];
static ultra_unregister_t *ultra_unregister_map[ 32 ];
static ultra_lseek_t *ultra_lseek_map[ 32 ];
static ultra_read_t *ultra_read_map[ 32 ];
static ultra_write_t *ultra_write_map[ 32 ];
static ultra_open_t *ultra_open_map[ 32 ];
static ultra_release_t *ultra_release_map[ 32 ];
#ifdef ULTRA_POLL
static ultra_poll_t *ultra_poll_map[ 32 ];
#else
static ultra_select_t *ultra_select_map[ 32 ];
#endif
static ultra_ioctl_t *ultra_ioctl_map[ 32 ];
static ultra_mmap_t *ultra_mmap_map[ 32 ];

#ifdef LINUX_2_1_X
static loff_t ultra_lseek( struct file *file, loff_t offset, int orig )
#else
#  ifdef LINUX_2_1
   static long long ultra_lseek( struct inode *inode, struct file *file, long long offset, int orig )
#  else
   static int ultra_lseek( struct inode *inode, struct file *file, off_t offset, int orig )
#  endif
#endif /* LINUX_2_1_X */
{
#ifdef LINUX_2_1_X
  unsigned short minor = MINOR( file -> f_dentry -> d_inode -> i_rdev );
#else
  unsigned short minor = MINOR( inode -> i_rdev );
#endif
  unsigned int idx;
  ultra_lseek_t *ptr;

  DEBUG_ACCESS( "lseek", minor );
  if ( minor >= ULTRA_MINOR_UBEGIN ) {
    idx = minor & ULTRA_MINOR_UDEVMASK;
  } else {
    idx = ULTRA_MINOR_OSS | ( minor & ULTRA_MINOR_OSS_MASK );
  }
  if ( VERIFY_CARD( idx, minor ) ) return -ENODEV;
  ptr = ultra_lseek_map[ idx ];
  if ( !ptr ) return -ENXIO;
  return ptr( file, offset, orig );
}    

#ifdef LINUX_2_1_X
static ssize_t ultra_read( struct file *file, char *buf, size_t count, loff_t *offset )
#else
#  ifdef LINUX_2_1
   static long ultra_read( struct inode *inode, struct file *file, char *buf, unsigned long count )
#  else
   static int ultra_read( struct inode *inode, struct file *file, char *buf, int count )
#  endif
#endif
{
#ifdef LINUX_2_1_X
  unsigned short minor = MINOR( file -> f_dentry -> d_inode -> i_rdev );
#else
  unsigned short minor = MINOR( inode -> i_rdev );
#endif
  unsigned int idx;
  ultra_read_t *ptr;

  DEBUG_ACCESS( "read", minor );
  if ( minor >= ULTRA_MINOR_UBEGIN ) {
    idx = minor & ULTRA_MINOR_UDEVMASK;
  } else {
    idx = ULTRA_MINOR_OSS | ( minor & ULTRA_MINOR_OSS_MASK );
  }
  if ( VERIFY_CARD( idx, minor ) ) return -ENODEV;
  ptr = ultra_read_map[ idx ];
  if ( !ptr ) return -ENXIO;
  return ptr( file, buf, count );
}

#ifdef LINUX_2_1_X
static ssize_t ultra_write( struct file *file, const char *buf, size_t count, loff_t *offset )
#else
#  ifdef LINUX_2_1
   static long ultra_write( struct inode *inode, struct file *file, const char *buf, unsigned long count )
#  else
   static int ultra_write( struct inode *inode, struct file *file, const char *buf, int count )
#  endif
#endif
{
#ifdef LINUX_2_1_X
  unsigned short minor = MINOR( file -> f_dentry -> d_inode -> i_rdev );
#else
  unsigned short minor = MINOR( inode -> i_rdev );
#endif
  unsigned int idx;
  ultra_write_t *ptr;

  DEBUG_ACCESS( "write", minor );
  if ( minor >= ULTRA_MINOR_UBEGIN ) {
    idx = minor & ULTRA_MINOR_UDEVMASK;
  } else {
    idx = ULTRA_MINOR_OSS | ( minor & ULTRA_MINOR_OSS_MASK );
  }
  if ( VERIFY_CARD( idx, minor ) ) return -ENODEV;
  ptr = ultra_write_map[ idx ];
  if ( !ptr ) return -ENXIO;
  return ptr( file, buf, count );
}

static int ultra_open( struct inode *inode, struct file *file )
{
  unsigned short minor = MINOR( inode -> i_rdev );
  unsigned int idx;
  ultra_open_t *ptr;

  DEBUG_ACCESS( "open", minor );
  if ( minor >= ULTRA_MINOR_UBEGIN ) {
    idx = minor & ULTRA_MINOR_UDEVMASK;
  } else {
    idx = ULTRA_MINOR_OSS | ( minor & ULTRA_MINOR_OSS_MASK );
  }
  if ( VERIFY_CARD( idx, minor ) ) return -ENODEV;
  ptr = ultra_open_map[ idx ];
  if ( !ptr ) return -ENXIO;
  return ptr( minor, file );
}

static int ultra_release( struct inode * inode, struct file *file )
{
  unsigned short minor = MINOR( inode -> i_rdev );
  unsigned int idx;
  ultra_release_t *ptr;

  DEBUG_ACCESS( "release", minor );
  if ( minor >= ULTRA_MINOR_UBEGIN ) {
    idx = minor & ULTRA_MINOR_UDEVMASK;
  } else {
    idx = ULTRA_MINOR_OSS | ( minor & ULTRA_MINOR_OSS_MASK );
  }
  if ( VERIFY_CARD( idx, minor ) ) return -ENODEV;
  ptr = ultra_release_map[ idx ];
  if ( !ptr ) return -ENXIO;
  return ptr( minor, file );
}

#ifdef ULTRA_POLL
static unsigned int ultra_poll( struct file *file, poll_table *wait )
{
#ifdef LINUX_2_1_X
  unsigned short minor = MINOR( file -> f_dentry -> d_inode -> i_rdev );
#else
  unsigned short minor = MINOR( file -> f_inode -> i_rdev );
#endif
  unsigned int idx;
  ultra_poll_t *ptr;

  DEBUG_ACCESS( "poll", minor );
  if ( minor >= ULTRA_MINOR_UBEGIN ) {
    idx = minor & ULTRA_MINOR_UDEVMASK;
  } else {
    idx = ULTRA_MINOR_OSS | ( minor & ULTRA_MINOR_OSS_MASK );
  }
  if ( VERIFY_CARD( idx, minor ) ) return -ENODEV;
  ptr = ultra_poll_map[ idx ];
  if ( !ptr ) return -ENXIO;
  return ptr( file, wait );
}
#else
static int ultra_select( struct inode *inode, struct file *file, 
                       int sel_type, select_table *wait )
{
  unsigned short minor = MINOR( inode -> i_rdev );
  unsigned int idx;
  ultra_select_t *ptr;

  DEBUG_ACCESS( "select", minor );
  if ( minor >= ULTRA_MINOR_UBEGIN ) {
    idx = minor & ULTRA_MINOR_UDEVMASK;
  } else {
    idx = ULTRA_MINOR_OSS | ( minor & ULTRA_MINOR_OSS_MASK );
  }
  if ( VERIFY_CARD( idx, minor ) ) return -ENODEV;
  ptr = ultra_select_map[ idx ];
  if ( !ptr ) return -ENXIO;
  return ptr( file, sel_type, wait );
}                       
#endif

static int ultra_ioctl( struct inode *inode, struct file *file,
		      unsigned int cmd, unsigned long arg )
{
  unsigned short minor = MINOR( inode -> i_rdev );
  unsigned int idx;
  ultra_ioctl_t *ptr;

  DEBUG_ACCESS( "ioctl", minor );
  if ( minor >= ULTRA_MINOR_UBEGIN ) {
    idx = minor & ULTRA_MINOR_UDEVMASK;
  } else {
    idx = ULTRA_MINOR_OSS | ( minor & ULTRA_MINOR_OSS_MASK );
  }
  if ( VERIFY_CARD( idx, minor ) ) return -ENODEV;
  ptr = ultra_ioctl_map[ idx ];
  if ( !ptr ) return -ENXIO;
  return ptr( file, cmd, arg );
}

#ifdef LINUX_2_1_X
static int ultra_mmap( struct file *file, struct vm_area_struct *vma )
#else
static int ultra_mmap( struct inode *inode, struct file *file, struct vm_area_struct *vma )
#endif
{
#ifdef LINUX_2_1_X
  struct inode *inode = file -> f_dentry -> d_inode;
  unsigned short minor = MINOR( inode -> i_rdev );
#else
  unsigned short minor = MINOR( file -> f_inode -> i_rdev );
#endif
  unsigned int idx;
  ultra_mmap_t *ptr;

  DEBUG_ACCESS( "mmap", minor );
  if ( minor >= ULTRA_MINOR_UBEGIN ) {
    idx = minor & ULTRA_MINOR_UDEVMASK;
  } else {
    idx = ULTRA_MINOR_OSS | ( minor & ULTRA_MINOR_OSS_MASK );
  }
  if ( VERIFY_CARD( idx, minor ) ) return -ENODEV;
  ptr = ultra_mmap_map[ idx ];
  if ( !ptr ) return -ENXIO;
  return ptr( inode, file, vma );
}

static struct file_operations ultra_fops = {
  ultra_lseek,	/* (l)lseek */
  ultra_read,	/* read */
  ultra_write,	/* write */
  NULL,		/* readdir */
#ifdef ULTRA_POLL
  ultra_poll,	/* poll */
#else
  ultra_select,	/* select */
#endif
  ultra_ioctl,	/* ioctl */
  ultra_mmap,	/* mmap */
  ultra_open,	/* open */
#ifdef LINUX_2_1
  ultra_release,	/* release */
#else
  (void (*)( struct inode * inode, struct file *file ))ultra_release,
#endif
#ifdef LINUX_2_1_X
  NULL,		/* fsync */
  NULL,		/* fasync */
  NULL,		/* check_media_change */
  NULL,		/* revalidate */
  NULL,		/* lock */
#endif
};

#ifndef LINUX_2_1

int ultra_ioctl_in( long *addr )
{
  if ( verify_area( VERIFY_READ, addr, sizeof( int ) ) != 0 )
    {
      printk( "ultra_ioctl_in: bad address 0x%lx\n", (long)addr );
      return 0;
    }
  return get_fs_int( addr );
}

int ultra_ioctl_out( long *addr, int value )
{
  int err;
  
  if ( ( err = verify_area( VERIFY_WRITE, addr, sizeof( int ) ) ) != 0 )
    return err;
  put_fs_int( value, addr );
  return 0;
}

#else

int ultra_ioctl_in( long *addr )
{
  int value;
  
  get_user( value, addr );
  return value;
}

int ultra_ioctl_out( long *addr, int value )
{
  int err;

  if ( ( err = verify_area( VERIFY_WRITE, addr, sizeof( int ) ) ) != 0 )
    return err;
  put_user( value, addr );
  return 0;
}

#endif /* !LINUX_2_1 */

static int ultra_fill_minor_map( int idx, struct ultra_stru_register_minor *reg )
{
  if ( reg ) {
    if ( ultra_unregister_map[ idx ] )	/* Oops.. Already registered!!! */
      return -EBUSY;
  } else {
    if ( !ultra_unregister_map[ idx ] ) /* Oops.. Already unregistered!!! */
      return 0;				/* Don't fail... */
  }

  if ( reg ) {
    ultra_comment_map[ idx ] = ultra_malloc_strdup( reg -> comment, SP_KERNEL );
  } else {
    ultra_free_str( ultra_comment_map[ idx ] );
    ultra_comment_map[ idx ] = NULL;
  }
  ultra_verify_map[ idx ] = reg ? reg -> verify_card : 0;
  ultra_unregister_map[ idx ] = reg ? reg -> ultra_unregister : NULL;
  ultra_lseek_map[ idx ] = reg ? reg -> ultra_lseek : NULL;
  ultra_read_map[ idx ] = reg ? reg -> ultra_read : NULL;
  ultra_write_map[ idx ] = reg ? reg -> ultra_write : NULL;
  ultra_open_map[ idx ] = reg ? reg -> ultra_open : NULL;
  ultra_release_map[ idx ] = reg ? reg -> ultra_release : NULL;
#ifdef ULTRA_POLL
  ultra_poll_map[ idx ] = reg ? reg -> ultra_poll : NULL;
#else
  ultra_select_map[ idx ] = reg ? reg -> ultra_select : NULL;
#endif
  ultra_ioctl_map[ idx ] = reg ? reg -> ultra_ioctl : NULL;
  ultra_mmap_map[ idx ] = reg ? reg -> ultra_mmap : NULL;
  return 0;
}

int ultra_register_minor( struct ultra_stru_register_minor *reg )
{
  int idx, idx1, minor, err;
  
  for ( idx = 0; idx < 16; idx++ ) {
    minor = reg -> ultra_minors[ idx ];
    if ( minor < 0 ) break;
    if ( ( err = ultra_fill_minor_map( minor, reg ) ) < 0 ) {
      /* Oops.. Try deallocate all minors.. */
      for ( idx1 = 0; idx1 < idx; idx1++ ) {
        minor = reg -> ultra_minors[ idx1 ];
        ultra_fill_minor_map( minor, NULL );
      }
      return err;
    }
  }

  for ( idx = 0; idx < 16; idx++ ) {
    minor = reg -> oss_minors[ idx ];
    if ( minor < 0 ) break;
    if ( ( err = ultra_fill_minor_map( ULTRA_MINOR_OSS | minor, reg ) ) < 0 ) {
      /* Oops.. Try deallocate all minors.. */
      for ( idx1 = 0; idx1 < idx; idx1++ ) {
        minor = reg -> oss_minors[ idx1 ];
        ultra_fill_minor_map( ULTRA_MINOR_OSS | minor, NULL );
      }
      for ( idx1 = 0; idx1 < 16; idx1++ ) {
        minor = reg -> ultra_minors[ idx1 ];
        if ( minor < 0 ) break;
        ultra_fill_minor_map( minor, NULL );
      }
      return err;
    }
  }

  return 0;
}

int ultra_unregister_minor( struct ultra_stru_register_minor *reg )
{
  int idx, minor;

  for ( idx = 0; idx < 16; idx++ ) {
    minor = reg -> ultra_minors[ idx ];
    if ( minor < 0 ) break;
    ultra_fill_minor_map( minor, NULL );
  }
  for ( idx = 0; idx < 16; idx++ ) {
    minor = reg -> oss_minors[ idx ];
    if ( minor < 0 ) break;
    ultra_fill_minor_map( ULTRA_MINOR_OSS | minor, NULL );
  }
  return 0;
}

/*
 *  INFO PART
 */

void ultra_minor_info( ultra_info_buffer_t *buffer )
{
  unsigned long flags;
  int idx;

  ultra_iprintf( buffer, "\nInstalled native devices:\n" );
  for ( idx = 0; idx < 16; idx++ ) {
    CLI( &flags );
    if ( !ultra_comment_map[ idx ] ) continue;
    ultra_iprintf( buffer, "  %02i [%s]: %s\n", idx, ultra_verify_map[ idx ] ? "card" : "glob", ultra_comment_map[ idx ] );
    STI( &flags );
  }
  ultra_iprintf( buffer, "Installed OSS devices:\n" );
  for ( idx = 16; idx < 32; idx++ ) {
    CLI( &flags );
    if ( !ultra_comment_map[ idx ] ) continue;
    ultra_iprintf( buffer, "  %02i [%s]: %s\n", idx - 16, ultra_verify_map[ idx ] ? "card" : "glob", ultra_comment_map[ idx ] );
    STI( &flags );
  }
}

/*
 *  INIT PART
 */

int init_module( void )
{
  MEMSET( &ultra_comment_map, 0, sizeof( ultra_comment_map ) );
  MEMSET( &ultra_verify_map, 0, sizeof( ultra_verify_map ) );
  MEMSET( &ultra_lseek_map, 0, sizeof( ultra_lseek_map ) );
  MEMSET( &ultra_read_map, 0, sizeof( ultra_read_map ) );
  MEMSET( &ultra_write_map, 0, sizeof( ultra_write_map ) );
  MEMSET( &ultra_open_map, 0, sizeof( ultra_open_map ) );
  MEMSET( &ultra_release_map, 0, sizeof( ultra_release_map ) );
#ifdef ULTRA_POLL
  MEMSET( &ultra_poll_map, 0, sizeof( ultra_poll_map ) );
#else
  MEMSET( &ultra_select_map, 0, sizeof( ultra_select_map ) );
#endif
  MEMSET( &ultra_ioctl_map, 0, sizeof( ultra_ioctl_map ) );
  MEMSET( &ultra_mmap_map, 0, sizeof( ultra_ioctl_map ) );

  if ( register_chrdev( ultra_major, "ultra", &ultra_fops ) ) {
    PRINTK("ultra: unable to register major device number %d\n", ultra_major );
    return -EIO;
  }
  ultra_malloc_init();
  ultra_driver_init();
  if ( ultra_info_register() < 0 ) {
    PRINTD( "ultra: unable to register info minors\n" );
    /* Not fatal error */
  }
  if ( ultra_control_register() < 0 ) {
    PRINTD( "ultra: unable to register control minors\n" );
    /* Not fatal error */
  }
  return 0;
}

void cleanup_module( void )
{
  int idx;
  ultra_unregister_t *unreg;
  
  for ( idx = 0; idx < 32; idx++ ) {
    unreg = ultra_unregister_map[ idx ];
    if ( unreg )
      unreg();
  }
  for ( idx = 0; idx < 32; idx++ ) {
    unreg = ultra_unregister_map[ idx ];
    if ( unreg )
      PRINTK( "ultra: %s device minor %i not unregistered!!!\n", idx & 0x10 ? "OSS" : "ULTRA", idx & 0x0f );
  }
  ultra_malloc_done();
  if ( unregister_chrdev( ultra_major, "ultra" ) != 0 )
    PRINTK( "ultra: unable to unregister major device number %d\n", ultra_major );
}
 